import os
import sys
import json
from glob import glob
from pwd import getpwnam
from subprocess import check_output

from charmhelpers.core.hookenv import (
    log, local_unit, status_set)

try:
    from landscape.client.configuration import (
        setup, register, LandscapeSetupConfiguration, ConfigurationError,
        stop_client_and_disable_init_script)
    from landscape.client.deployment import Configuration
    from landscape.client.sysvconfig import SysVConfig
except ImportError:
    # Compatibility with landscape-clients older than 18.01 (pre bionic)
    from landscape.configuration import (
        setup, register, LandscapeSetupConfiguration, ConfigurationError,
        stop_client_and_disable_init_script)
    from landscape.deployment import Configuration
    from landscape.sysvconfig import SysVConfig


class JujuBroker(object):
    """Class responsible for communicating with Juju.

    This class should be used to set and get information from Juju, so
    that it can be overridden in unit tests, where a complete Juju
    environment isn't available.
    """

    def __init__(self):
        self.environment = os.environ.copy()

    def get_relation_config(self):
        """Get relation configuration from Juju."""
        return self._run_juju_tool("relation-get")

    def get_machine_id(self):
        """Return the Juju machine ID of this unit."""
        # XXX once #1359714 is fixed this method can be dropped and we can use
        # the JUJU_MACHINE_ID environment variable.

        # Sniff the value of the machine ID by looking at the name of the
        # directory where hooks live. It will be something like
        # 'machine-0-lxc-1'.
        pattern = "../../machine-*"
        match = glob(pattern)[0]
        dirname = os.path.basename(match)
        return dirname.lstrip("machine-").replace("-", "/")

    def get_service_config(self):
        """Get service configuration from Juju."""
        return self._run_juju_tool("config-get")

    def log(self, message):
        """Log a message via Juju."""
        # --format is deprecated for juju-log
        return check_output(["juju-log", message])

    def _run_juju_tool(self, command):
        """Run specified Juju tool and parse json output."""
        output = check_output([command, "--format", "json"])
        return json.loads(output) if output else {}


class LandscapeBroker(object):
    """Class responsible for communicating with the Landscape client.

    This class should be used to interact with Landscape client, so
    that it can be overridden in unit tests, where we can't let
    Landscape access system-specific behaviours.
    """

    def __init__(self):
        self.config = Configuration()
        self.status_set = status_set  # For testing

    @property
    def local_unit(self):
        return local_unit()

    def update_client_config(self, service_config, try_to_register=True):
        """Update the client config.

        @param service_config: A dict containing the config items to be
            updated. Dashes in the key values will be converted to
            underscore, to match the Landscape configuration.

        @param try_to_register: Should we attempt to register after
            updating the client configuration? (default True). No need
            to show warnings if we are updating the configuration to
            be deleted (as is the case when exiting a relation, for
            example).

        An attempt to register the client will be made, if the client wasn't
        configured before.
        """
        self.config.reload()
        for key, value in service_config.items():
            if key != "origin":
                setattr(self.config, key.replace("-", "_"), value)
        self.config.write()
        is_configured_enough = all([
            self.config.get("account_name"),
            self.config.get("computer_title"),
            os.path.exists(self.juju_filename)])
        if not try_to_register:
            # We don't need to further check the configuration if we are
            # deleting it.
            return 0

        if is_configured_enough:
            exit_code = self.try_to_register()
            if exit_code == 0:
                self.status_set("active", "System successfully registered")
        else:
            if not self.config.get("account_name"):
                self.status_set(
                    "blocked", "Need account-name/registration-key to proceed")
            else:
                self.status_set(
                    "maintenance",
                    "Need computer-title and juju-info to proceed")
            return 0

    @property
    def juju_filename(self):
        try:
            return self.config.juju_filename
        except AttributeError:  # older landscape-client versions
            return os.path.join(self.config.data_path, "juju-info.json")

    def clear_registration(self):
        """
        Do steps necessary to clear the client registration, so that next time
        it will re-register
        """
        stop_client_and_disable_init_script()

    def try_to_register(self):
        """Try to register the client if needed.

        If the client has already been configured, no registration attempt
        will be made. The same is true if a computer title hasn't been set,
        since that means that the juju-info-relation-joined hasn't been run,
        which means that it's expected that the configuration will be broken
        at that point.

        The following keys need to be set for an attempt to be made:
         - account_name, computer_title

        If an error is encountered, the client will be disabled and the hook
        calling this function will exit with an error code.
        """
        log("Updating client configuration for %s" % self.local_unit)
        error_handler = ErrorHandler()

        sysvconfig = SysVConfig()
        configured = sysvconfig.is_configured_to_run()
        if configured:
            log("%s: landscape-client is configured to run as per sysvconfig"
                "" % self.local_unit)
        else:
            log("%s: landscape-client is not configured to run as per "
                "sysvconfig" % self.local_unit)

        config = LandscapeSetupConfiguration()
        config.load([])
        config.silent = True
        if configured:
            print("Client already registered, skipping")
            return 0
        try:
            setup(config)
        except ConfigurationError as error:
            print >> sys.stderr, "Configuration error: %s" % (str(error),)
            error_handler.flag_error(1)
        else:
            log("Registering %s" % self.local_unit)
            register(config, on_error=error_handler.flag_error)
        return error_handler.exit_code


class ErrorHandler(object):

    exit_code = 0

    def flag_error(self, exit_code):
        """Disable the client and indicate an error was encountered.

        The client is disabled, so that the config-changed hook can be
        re-run after the configuration error has been resolve. If the client
        wouldn't be disabled, try_to_register() would see that the client is
        configured and wouldn't attempt to register the client again.
        """
        # Disable the client so that the config-change hook can be re-run
        # after the error has been resolved.
        stop_client_and_disable_init_script()
        self.exit_code = exit_code


def chown(filename):
    """Change file owner and group to landscape."""
    user = getpwnam("landscape")
    os.chown(filename, user.pw_uid, user.pw_gid)


def write_json_file(filename, content):
    """Dump content in JSON format to the specified file."""
    with open(filename, "w") as dumpfile:
        json.dump(content, dumpfile)
