import apt_pkg
import sys
import socket
import os
import base64

import six
from charmhelpers.core.hookenv import (
    Hooks, UnregisteredHookError, log)
from shutil import rmtree
from subprocess import CalledProcessError

from charmhelpers import fetch
from charmhelpers.core import hookenv

from common import (
    write_json_file, JujuBroker, LandscapeBroker, chown)
from ceph import (
    get_ceph_client_path, write_ceph_client_keyring, write_ceph_client_config)
from install import install_landscape_client


CERT_FILE = "/etc/ssl/certs/landscape_server_ca.crt"

JUJU_BROKER = JujuBroker()
LANDSCAPE_BROKER = LandscapeBroker()

hooks = Hooks()


@hooks.hook("registration-relation-joined", "registration-relation-changed")
def registration_relation(juju_broker=JUJU_BROKER,
                          hookenv=hookenv, fetch=fetch,
                          landscape_broker=LANDSCAPE_BROKER):
    data = juju_broker.get_relation_config()
    if "ping-url" in data:
        config_changed(relation_data=data, hookenv=hookenv, fetch=fetch,
                       landscape_broker=landscape_broker)


@hooks.hook("config-changed")
def config_changed(relation_data=None, hookenv=hookenv, fetch=fetch,
                   landscape_broker=LANDSCAPE_BROKER):
    """
    @param relation_data: data from the context of a relation, it should
                          match the data in the config
    """
    hookenv.log("In config-changed for %s" % landscape_broker.local_unit)
    if relation_data is None:
        relation_data = {}
    service_config = hookenv.config()
    if service_config.changed("origin"):
        install_landscape_client(service_config["origin"], fetch=fetch)
    # Only update client config with service configs that have been set
    set_config = {}
    for key, value in six.iteritems(service_config):
        if value is not None:
            if type(value) == str:
                set_config[key] = value.strip()
            else:
                set_config[key] = value

    relation_data.update(set_config)
    if service_config.get("disable-unattended-upgrades"):
        Cnf = apt_pkg.Configuration()
        apt_dir = '/etc/apt/apt.conf.d'
        apt_pkg.read_config_dir(Cnf, apt_dir)
        if Cnf.find("APT::Periodic::Unattended-Upgrade") == '1':
            disable_upgrades = 'APT::Periodic::Unattended-Upgrade "0";'
            hookenv.log("Disabling unattended upgrades")
            with open(apt_dir + '/99landscapeoverride', 'w') as override_file:
                override_file.write(disable_upgrades)
    if relation_data.get("ssl-public-key", "").startswith("base64:"):
        _write_certificate(relation_data["ssl-public-key"][7:], CERT_FILE)
        relation_data["ssl-public-key"] = CERT_FILE
    if "account-name" in relation_data:
        return landscape_broker.update_client_config(relation_data)


@hooks.hook("upgrade-charm")
def upgrade_charm(landscape_broker=LANDSCAPE_BROKER):
    """Idempotently upgrades state from older charms."""
    landscape_broker.config.reload()
    return migrate_old_juju_info(landscape_broker.config)


@hooks.hook("container-relation-joined", "ceph-client-relation-joined")
def container_relation_joined(juju_broker=JUJU_BROKER,
                              landscape_broker=LANDSCAPE_BROKER):
    juju_broker.log("In container-relation-joined for %s"
                    "" % landscape_broker.local_unit)
    landscape_broker.config.reload()
    # We use the remote unit for the unit name, since we want to associate this
    # client with the unit it's managing, not its own unit.
    model_uuid = juju_broker.environment.get(
        "JUJU_MODEL_UUID", juju_broker.environment.get("JUJU_ENV_UUID"))
    juju_info = {
        "environment-uuid": model_uuid,
        "machine-id": juju_broker.get_machine_id(),
        "api-addresses": juju_broker.environment.get("JUJU_API_ADDRESSES")}

    write_json_file(landscape_broker.juju_filename, juju_info)
    exit_code = landscape_broker.update_client_config(
        {"computer-title": socket.gethostname()})

    return exit_code


@hooks.hook("container-relation-departed")
def container_relation_departed(landscape_broker=LANDSCAPE_BROKER):
    """Stop managing the removed unit with Landscape.

    If the relation is removed, the registration will be cleared, the
    computer title unset and the juju info file removed, so that we
    won't try to register again when the config-changed hook is fired.
    """
    landscape_broker.clear_registration()

    # Delete all of the juju-info files.
    if os.path.exists(landscape_broker.juju_filename):
        os.remove(landscape_broker.juju_filename)
    landscape_broker.update_client_config(
        {"computer-title": ""}, try_to_register=False)


@hooks.hook("ceph-client-relation-changed")
def ceph_relation_changed(juju_broker=JUJU_BROKER,
                          landscape_broker=LANDSCAPE_BROKER):
    relation_config = juju_broker.get_relation_config()
    # The key is only reported once the ceph ring has quorum (there are enough
    # nodes in the service). There's nothing to do until the key is provided.
    if "key" not in relation_config:
        juju_broker.log(
            "Ceph key not reported in relation config, nothing to do.")
        return 0

    ceph_client_dir = get_ceph_client_path(landscape_broker)
    if not os.path.exists(ceph_client_dir):
        os.mkdir(ceph_client_dir)
        chown(ceph_client_dir)

    # Create ceph client keyring and config for landscape-client use.
    try:
        keyring_file = write_ceph_client_keyring(
            landscape_broker, "landscape-client", relation_config["key"])
    except CalledProcessError as err:
        juju_broker.log(
            "Writing ceph keyring file failed with code %d" % err.returncode)
        return 1

    try:
        write_ceph_client_config(
            landscape_broker, relation_config["auth"], keyring_file,
            relation_config["private-address"])
    except IOError as err:
        juju_broker.log("Writing ceph configuration file failed: %s" % err)
        return 1
    else:
        juju_broker.log("Ceph configuration written to file.")
        return 0


@hooks.hook("ceph-client-relation-broken")
def ceph_relation_broken(landscape_broker=LANDSCAPE_BROKER):
    rmtree(get_ceph_client_path(landscape_broker), ignore_errors=True)


@hooks.hook("ceph-client-relation-departed")
def ceph_relation_departed(landscape_broker=LANDSCAPE_BROKER):
    landscape_broker.clear_registration()


def migrate_old_juju_info(client_config):
    """
    Migrates data from the old meta-data.d directory into the new
    juju-info.json file.
    """
    fallback_meta_data_dir = os.path.join(
        client_config.data_path, "meta-data.d")
    meta_data_dir = client_config.get("meta_data_path", fallback_meta_data_dir)
    if not os.path.exists(meta_data_dir):
        return 0
    juju_info_file = os.path.join(client_config.data_path, "juju-info.json")
    juju_key_map = {"juju-env-uuid": "environment-uuid",
                    "juju-unit-name": "unit-name",
                    "juju-api-addresses": "api-addresses",
                    "juju-private-address": None}
    juju_info = dict.fromkeys(key for key in juju_key_map.values() if key)
    files_to_delete = []
    for filename, info_key in six.iteritems(juju_key_map):
        file_path = os.path.join(meta_data_dir, filename)
        if not os.path.exists(file_path):
            continue
        files_to_delete.append(file_path)
        if info_key is None:
            continue
        with open(file_path, "r") as juju_info_part_file:
            juju_info[info_key] = juju_info_part_file.read().strip()
    missing_values = [value is None for value in juju_info.itervalues()]
    # If they're all missing it probably means we've already converted.
    if all(missing_values):
        return 0
    # ... otherwise there's something unexpected going on.
    if any(missing_values):
        return "Incomplete Juju information found in %s" % meta_data_dir
    write_json_file(juju_info_file, juju_info)
    for file_to_delete in files_to_delete:
        os.remove(file_to_delete)
    return 0


def _write_certificate(certificate, filename):
    """
    @param certificate Text of the certificate, base64 encoded.
    @param filename Full path to file to write
    """
    with open(filename, "wb") as file:
        file.write(base64.b64decode(certificate))


if __name__ == '__main__':
    try:
        sys.exit(hooks.execute(sys.argv[1:]))
    except UnregisteredHookError as e:
        log('Unknown hook {} - skipping.'.format(e))
