#!/usr/bin/python3
"""This is the install hook file.

It is its own file because the rest of the hooks import landscape classes and
functions, and landscape is not installed until this hook runs.
"""

import os
import subprocess
import sys

from glob import glob

from charmhelpers.core.hookenv import (
    Hooks, UnregisteredHookError, log)
from charmhelpers import fetch
from charmhelpers.core.hookenv import config


hooks = Hooks()


def install_landscape_client(origin, fetch=fetch):
    """Update the origin and install/upgrade landscape-client."""
    if origin not in ["distro", None]:
        add_apt_source(origin, fetch=fetch)

    fetch.apt_update()
    fetch.apt_install(["landscape-client"])


@hooks.hook("install")
def install():
    charm_config = config()
    fetch.apt_install(["apt-transport-https", "wget"])

    install_landscape_client(charm_config.get("origin"))

    data_path = charm_config.get("data-path")
    if not data_path:
        data_path = "/var/lib/landscape/client"

    result = subprocess.call(["landscape-config", "--init", "-d", data_path])
    if result != 0:
        result = subprocess.check_call([
            "install", "-o", "landscape", "-g", "root", "-m", "755", "-d",
            data_path])


def build_from_launchpad(url):
    """The charm will install the code from the passed lp branch.
    """
    fetch.apt_install(["devscripts", "bzr", "pbuilder"], fatal=True)
    subprocess.check_call(["rm", "-rf", "landscape-client-source"])
    subprocess.check_call(["bzr", "branch", url, "landscape-client-source"])
    os.chdir("landscape-client-source")
    subprocess.check_call("/usr/lib/pbuilder/pbuilder-satisfydepends")
    env = {"DEBUILD_OPTS": "-uc -us"}
    subprocess.check_call(["make", "package"], env=env)
    # TODO: The following call should be retried (potential race condition to
    # acquire the dpkg lock).
    subprocess.call(["dpkg", "-i", glob("../landscape-client_*.deb")[0],
                     glob("../landscape-common_*.deb")[0]])
    # The _run_with_retries will ensure the command is retried in case we can't
    # acquire the lock for some reason.
    fetch._run_apt_command(["apt-get", "-f", "install"], fatal=True)
    os.chdir("..")


def add_apt_source(url, fetch=fetch):
    """Add an apt source entry, with passed in key.

    This is a thin wrapper over the charmhelper "add_source" method to allow
    for the URLs to contain the key explicitely, like:
        - https://blah/blah|KEY
        - deb https://blah/blah|KEY
        - ppa:landscape/blah
    """
    key = None
    if url.startswith("lp"):
        return build_from_launchpad(url)

    if "|" in url:
        # We have a particular key specified. Url = url without the key now.
        url, key = url.split("|", 1)  # Maximum of 1 split.
        if key == "":
            log("Archive key for '%s' is empty.", level="WARNING")

    return fetch.add_source(url, key)


if __name__ == '__main__':
    try:
        sys.exit(hooks.execute(sys.argv))
    except UnregisteredHookError as e:
        log('Unknown hook {} - skipping.'.format(e))
