import json
import socket
import os
from shutil import rmtree
from tempfile import mkdtemp
from unittest import TestCase
from uuid import uuid4

import six
from landscape.deployment import Configuration
from landscape.lib.fs import read_file
import hooks
from common import LandscapeBroker, JujuBroker, write_json_file
from stubs import FetchStub, HookenvStub


class FakeJujuBroker(JujuBroker):
    """A JujuBroker that can be used in unit tests.

    Instead of executing commands, it returns the output from
    self.commands, which tests can set to the required output.
    """

    def __init__(self):
        self.commands = {"relation-get": {}, "config-get": {}}
        self.environment = {
            "JUJU_ENV_UUID": str(uuid4()),
            "JUJU_API_ADDRESSES": "juju.example.com:12345"}
        self.logs = []

    def log(self, message):
        self.logs.append(message)

    def get_machine_id(self):
        return "1"

    def _run_juju_tool(self, command):
        """Override _run_juju_tool not not execute any commands."""
        return self.commands[command]


class FakeLandscapeBroker(LandscapeBroker):
    """A LandscapeBroker that can be used in unittests."""

    cleared_registration = False

    def try_to_register(self):
        """Don't check system files or try to actually register the client."""
        return False

    def clear_registration(self):
        self.cleared_registration = True

    @property
    def local_unit(self):
        return "test-unit"


class HookTestCase(TestCase):
    """Base test class for testing hooks."""

    def setUp(self):
        data_path = self.mkdtemp()
        self.landscape_broker = FakeLandscapeBroker()
        self.landscape_broker.config.data_path = data_path
        self.landscape_broker.config.config = os.path.join(
            data_path, "landscape.conf")
        self.landscape_broker.config.write()
        self.workload_state = None
        self.workload_message = None
        self.charm_dir = self.mkdtemp()
        os.environ["CHARM_DIR"] = self.charm_dir
        self.hookenv = HookenvStub(self.charm_dir)
        self.fetch = FetchStub(needs_update=True)

        def status_set(workload_state, message):
            self.workload_state = workload_state
            self.workload_message = message

        self.landscape_broker.status_set = status_set
        self.juju_broker = FakeJujuBroker()

    def touch_juju_info_file(self):
        write_json_file(self.landscape_broker.juju_filename, content={})

    def mkdtemp(self):
        directory = mkdtemp(prefix=self.__class__.__name__)
        self.addCleanup(rmtree, directory)
        return directory

    def config(self):
        """Return the base application config."""
        config = self.hookenv.config()
        config.update({"origin": "distro"})
        return config


class MigrateOldJujuInfoTest(HookTestCase):

    def setUp(self):
        self.config = Configuration()
        self.config.data_path = self.mkdtemp()
        self.config.meta_data_path = os.path.join(
            self.config.data_path, "metadata.d")
        os.makedirs(self.config.meta_data_path)
        super(MigrateOldJujuInfoTest, self).setUp()

    def setup_old_juju_info(self, juju_info):
        for filename, contents in six.iteritems(juju_info):
            juju_path = os.path.join(self.config.meta_data_path, filename)
            with open(juju_path, "w") as juju_file:
                juju_file.write(contents)

    def test_successful_migration(self):
        """
        Juju info in an old metadata.d directory is converted into a
        single JSON file.
        """
        juju_info = {"juju-env-uuid": "0afcd28d-6263-43fb-8131-afa696a984f8",
                     "juju-machine-id": "1",
                     "juju-unit-name": "postgresql/0",
                     "juju-api-addresses": "10.0.3.1:17070",
                     "juju-private-address": "10.0.3.205"}
        self.setup_old_juju_info(juju_info)
        hooks.migrate_old_juju_info(self.config)
        juju_info_path = os.path.join(self.config.data_path, "juju-info.json")
        self.assertTrue(os.path.exists(juju_info_path))
        with open(juju_info_path) as juju_info_fh:
            juju_json_info = json.load(juju_info_fh)
        self.assertEqual(
            {"environment-uuid": "0afcd28d-6263-43fb-8131-afa696a984f8",
             "unit-name": "postgresql/0",
             "api-addresses": "10.0.3.1:17070"}, juju_json_info)

    def test_old_juju_info_deleted_after_successful_migration(self):
        """
        Old Juju files are deleted when the information is succesfully
        converted.
        """
        juju_info = {"juju-env-uuid": "0afcd28d-6263-43fb-8131-afa696a984f8",
                     "juju-unit-name": "postgresql/0",
                     "juju-api-addresses": "10.0.3.1:17070",
                     "juju-private-address": "10.0.3.205"}
        self.setup_old_juju_info(juju_info)
        hooks.migrate_old_juju_info(self.config)
        self.assertEqual([], os.listdir(self.config.meta_data_path))

    def test_incomplete_juju_info_is_error(self):
        """
        If only part of the Juju info is in meta-data.d directory then the
        function returns non-zero and the files aren't deleted.
        """
        juju_info = {"juju-env-uuid": "0afcd28d-6263-43fb-8131-afa696a984f8",
                     "juju-api-addresses": "10.0.3.1:17070",
                     "juju-private-address": "10.0.3.205"}
        self.setup_old_juju_info(juju_info)
        self.assertNotEqual(0, hooks.migrate_old_juju_info(self.config))
        self.assertGreater(len(os.listdir(self.config.meta_data_path)), 0)

    def test_idempotent(self):
        """
        Running the upgrade code more than once has no effect and is not
        an error.
        """
        juju_info = {"juju-env-uuid": "0afcd28d-6263-43fb-8131-afa696a984f8",
                     "juju-unit-name": "postgresql/0",
                     "juju-api-addresses": "10.0.3.1:17070",
                     "juju-private-address": "10.0.3.205"}
        self.setup_old_juju_info(juju_info)
        hooks.migrate_old_juju_info(self.config)
        self.assertEqual(0, hooks.migrate_old_juju_info(self.config))


class ContainerRelationJoinedTest(HookTestCase):
    """Test for the container-relation-joined hook."""

    def setUp(self):
        super(ContainerRelationJoinedTest, self).setUp()
        self.juju_broker.environment["JUJU_REMOTE_UNIT"] = "remote/0"
        self.landscape_broker.config.account_name = "account1"
        self.landscape_broker.config.write()

    def test_writes_juju_1_json(self):
        """
        When the hook runs in Juju 1, the environment UUID, which unit the
        client is managed, the Juju API addresses, and the unit's
        private address are written to the juju-info.json file inside
        the client's data path.
        """
        self.juju_broker.environment.update({
            "JUJU_ENV_UUID": "uuid1",
            "JUJU_API_ADDRESSES": "10.0.0.1:1234"})
        hooks.container_relation_joined(juju_broker=self.juju_broker,
                                        landscape_broker=self.landscape_broker)

        juju_info_path = self.landscape_broker.juju_filename
        juju_info = json.loads(read_file(juju_info_path))
        self.assertEqual(
            {u"environment-uuid": u"uuid1",
             u"machine-id": u"1",
             u"api-addresses": u"10.0.0.1:1234"},
            juju_info)

    def test_writes_juju_2_json(self):
        """
        When the hook runs in Juju 2, the model UUID, which unit the
        client is managed, the Juju API addresses, and the unit's
        private address are written to the juju-info.json file inside
        the client's data path.
        """
        self.juju_broker.environment.update({
            "JUJU_MODEL_UUID": "uuid2",
            "JUJU_API_ADDRESSES": "10.0.0.1:1234"})
        hooks.container_relation_joined(juju_broker=self.juju_broker,
                                        landscape_broker=self.landscape_broker)

        juju_info_path = self.landscape_broker.juju_filename
        juju_info = json.loads(read_file(juju_info_path))
        self.assertEqual(
            {u"environment-uuid": u"uuid2",
             u"machine-id": u"1",
             u"api-addresses": u"10.0.0.1:1234"},
            juju_info)

    def test_registers_after_juju_info_is_written(self):
        """
        The client tries to register to Landscape after the juju-info
        file has been written, so that juju information is sent in the
        registration message.
        """
        registration_attempts = []

        def record_registration_attempt():
            if os.path.exists(self.landscape_broker.juju_filename):
                registration_attempts.append(True)

        self.landscape_broker.try_to_register = record_registration_attempt
        hooks.container_relation_joined(juju_broker=self.juju_broker,
                                        landscape_broker=self.landscape_broker)

        self.assertEqual([True], registration_attempts)

    def test_computer_title_is_hostname(self):
        """
        The client adds the hostname as the "computer-title" field of the
        landscape broker config.
        """
        hooks.container_relation_joined(juju_broker=self.juju_broker,
                                        landscape_broker=self.landscape_broker)
        result = self.landscape_broker.config.computer_title
        self.assertEqual(result, socket.gethostname())


class ContainerRelationDepartedTest(HookTestCase):
    """Test for the container-relation-departed hook."""

    def test_clears_registration(self):
        """
        When the relation is removed, the Landscape registration data
        is removed.
        """
        hooks.container_relation_departed(
            landscape_broker=self.landscape_broker)
        self.assertTrue(self.landscape_broker.cleared_registration)

    def test_clears_computer_title(self):
        """
        When the relation is removed, the computer title is cleared in
        the Landscape config, since it's the name of the remove unit
        (which is now gone).
        """
        self.landscape_broker.config.computer_title = "unit/0"
        hooks.container_relation_departed(
            landscape_broker=self.landscape_broker)
        self.assertEqual("", self.landscape_broker.config.computer_title)

    def test_removes_juju_info(self):
        """
        When the relation is removed, the juju info file is removed,
        since it contains information about the remove unit (which is
        now gone).
        """
        juju_info_path = self.landscape_broker.juju_filename
        juju_info = {"environment-uuid": "uuid1",
                     "api-addresses": "10.0.0.1:1234"}
        write_json_file(juju_info_path, juju_info)
        hooks.container_relation_departed(
            landscape_broker=self.landscape_broker)
        self.assertFalse(os.path.exists(juju_info_path))


class CephClientRelationJoinedTest(HookTestCase):
    """Test for the ceph-client-relation-changed hook."""

    def setUp(self):
        super(CephClientRelationJoinedTest, self).setUp()
        self.addCleanup(setattr, hooks, "chown", hooks.chown)
        self.addCleanup(
            setattr, hooks, "write_ceph_client_keyring",
            hooks.write_ceph_client_keyring)
        self.addCleanup(
            setattr, hooks, "write_ceph_client_config",
            hooks.write_ceph_client_config)

        self.chowned_files = []

        def stub_chown(filepath):
            self.chowned_files.append(filepath)
        hooks.chown = stub_chown

    def test_nothing_to_do_with_no_key(self):
        """
        If there's no key in the relation information provided by ceph,
        log and do nothing.
        """
        self.assertEqual(0, hooks.ceph_relation_changed(
            juju_broker=self.juju_broker,
            landscape_broker=self.landscape_broker))

        [log] = self.juju_broker.logs
        self.assertEqual(
            "Ceph key not reported in relation config, nothing to do.", log)

    def test_ceph_client_path_made(self):
        """
        If the ceph client path doesn't exist already it's created and chowned.
        """
        self.juju_broker.commands["relation-get"] = {
            "key": "foo", "auth": "bar", "private-address": "home"}

        def stub_ceph_write(*args):
            pass
        hooks.write_ceph_client_keyring = stub_ceph_write
        hooks.write_ceph_client_config = stub_ceph_write

        self.assertEqual(0, hooks.ceph_relation_changed(
            juju_broker=self.juju_broker,
            landscape_broker=self.landscape_broker))

        ceph_client_dir = os.path.join(
            self.landscape_broker.config.data_path, "ceph-client")
        self.assertTrue(os.path.exists(ceph_client_dir))
        self.assertEqual([ceph_client_dir], self.chowned_files)

    def test_write_ceph_client_keyring_called(self):
        """
        write_ceph_client_keyring is called with the key from the Juju
        relation config.
        """
        self.juju_broker.commands["relation-get"] = {
            "key": "foo", "auth": "bar", "private-address": "home"}

        called_args = []

        def stub_write_ceph_client_keyring(*args):
            called_args.append(args)

        def stub_ceph_write(*args):
            pass

        hooks.write_ceph_client_keyring = stub_write_ceph_client_keyring
        hooks.write_ceph_client_config = stub_ceph_write

        self.assertEqual(0, hooks.ceph_relation_changed(
            juju_broker=self.juju_broker,
            landscape_broker=self.landscape_broker))
        [args] = called_args
        self.assertEqual(
            (self.landscape_broker, "landscape-client", "foo"), args)

    def test_write_ceph_client_config_called(self):
        """
        write_ceph_client_config is called with the auth and
        private-address from the Juju relation config.
        """
        self.juju_broker.commands["relation-get"] = {
            "key": "foo", "auth": "bar", "private-address": "home"}

        called_args = []
        keyring_file = object()

        def stub_ceph_write(*args):
            return keyring_file

        hooks.write_ceph_client_keyring = stub_ceph_write

        def stub_write_ceph_client_config(*args):
            called_args.append(args)
        hooks.write_ceph_client_config = stub_write_ceph_client_config

        self.assertEqual(0, hooks.ceph_relation_changed(
            juju_broker=self.juju_broker,
            landscape_broker=self.landscape_broker))
        [args] = called_args
        self.assertEqual(
            (self.landscape_broker, "bar", keyring_file, "home"), args)

    def test_confirmation_logged(self):
        self.juju_broker.commands["relation-get"] = {
            "key": "foo", "auth": "bar", "private-address": "home"}

        def stub_ceph_write(*args):
            pass
        hooks.write_ceph_client_keyring = stub_ceph_write
        hooks.write_ceph_client_config = stub_ceph_write

        self.assertEqual(0, hooks.ceph_relation_changed(
            juju_broker=self.juju_broker,
            landscape_broker=self.landscape_broker))

        [log] = self.juju_broker.logs
        self.assertEqual("Ceph configuration written to file.", log)


class CephClientRelationBrokenTest(HookTestCase):
    """Test for the ceph-client-relation-broken hook."""

    def test_ceph_config_dir_removed(self):
        ceph_client_path = os.path.join(
            self.landscape_broker.config.data_path, "ceph-client")
        os.mkdir(ceph_client_path)
        hooks.ceph_relation_broken(landscape_broker=self.landscape_broker)
        self.assertFalse(os.path.exists(ceph_client_path))


class ConfigChangedTest(HookTestCase):
    """Test for the config-changed hook."""

    def setUp(self):
        super(ConfigChangedTest, self).setUp()
        self.landscape_broker.config.account_name = "account1"
        self.landscape_broker.config.computer_title = "title"
        self.addCleanup(setattr, hooks, "CERT_FILE", hooks.CERT_FILE)
        tmp_dir = self.mkdtemp()
        hooks.CERT_FILE = "%s/cert" % tmp_dir

    def test_ssl_cert_decoded(self):
        """
        A base64 encoded SSL cert in the configuration is decoded.
        """
        self.touch_juju_info_file()
        config = self.config()
        config.update({
            "ssl-public-key": "base64:%s" % "foo".encode("base64"),
            "account-name": "standalone",
            "ping-url": "http://127.0.0.1"})

        hooks.config_changed(hookenv=self.hookenv, fetch=self.fetch,
                             landscape_broker=self.landscape_broker)
        self.assertEqual(
            hooks.CERT_FILE, self.landscape_broker.config.ssl_public_key)
        with open(hooks.CERT_FILE) as fh:
            self.assertEqual("foo", fh.read())
        self.assertEqual("active", self.workload_state)
        self.assertEqual(
            "System successfully registered", self.workload_message)

    def test_no_account_name(self):
        """
        If account-name is not set, the unit transitions to the blocked state.
        """
        self.touch_juju_info_file()
        config = self.config()
        config.update({
            "ssl-public-key": "base64:%s" % "foo".encode("base64"),
            "account-name": "",
            "ping-url": "http://127.0.0.1"})

        hooks.config_changed(hookenv=self.hookenv, fetch=self.fetch,
                             landscape_broker=self.landscape_broker)
        self.assertEqual("blocked", self.workload_state)
        self.assertEqual(
            "Need account-name/registration-key to proceed",
            self.workload_message)

    def test_no_juju_info(self):
        """
        If the juju-info file hasn't been written yet, the unit transitions to
        the maintenance status.
        """
        config = self.config()
        config.update({
            "ssl-public-key": "base64:%s" % "foo".encode("base64"),
            "account-name": "standalone",
            "ping-url": "http://127.0.0.1"})

        hooks.config_changed(hookenv=self.hookenv, fetch=self.fetch,
                             landscape_broker=self.landscape_broker)
        self.assertEqual("maintenance", self.workload_state)
        self.assertEqual(
            "Need computer-title and juju-info to proceed",
            self.workload_message)

    def test_whitespace_removed_from_config_settings(self):
        """Leading and trailing whitespace is removed from config settings"""
        config = self.config()
        config.update({
            "ssl-public-key": " base64:%s" % "foo".encode("base64"),
            "account-name": " standalone\t",
            "registration-key": " mykey\n",
            "ping-url": " \thttp://127.0.0.1\n "})
        hooks.config_changed(hookenv=self.hookenv, fetch=self.fetch,
                             landscape_broker=self.landscape_broker)
        self.assertEqual(
            hooks.CERT_FILE, self.landscape_broker.config.ssl_public_key)
        with open(hooks.CERT_FILE) as fh:
            self.assertEqual("foo", fh.read())
        self.assertEqual(
            "standalone", self.landscape_broker.config.account_name)
        self.assertEqual(
            "mykey", self.landscape_broker.config.registration_key)
        self.assertEqual(
            "http://127.0.0.1", self.landscape_broker.config.ping_url)

    def test_update_origin(self):
        """"
        If origin is set, the origin will be added as an apt source and
        landscape-client will be installed.
        """
        config = self.config()
        config.update({"origin": "ppa:landscape-client/stable"})
        hooks.config_changed(hookenv=self.hookenv, fetch=self.fetch,
                             landscape_broker=self.landscape_broker)
        self.assertEqual(
            [("ppa:landscape-client/stable", None)], self.fetch.sources)
        [install] = self.fetch.installs
        self.assertEqual(["landscape-client"], install.packages)

    def test_update_origin_twice(self):
        """"
        If origin hasn't been changed since the last time the
        config-changed hook ran, the origin won't be added again, and
        landscape-client won't be installed again.
        """
        config = self.config()
        config.update({"origin": "ppa:landscape-client/stable"})
        hooks.config_changed(hookenv=self.hookenv, fetch=self.fetch,
                             landscape_broker=self.landscape_broker)
        # Saving and loading the config is taken care of by
        # Hooks.execute(), which we can't use in the test.
        config.save()
        config.load_previous()
        hooks.config_changed(hookenv=self.hookenv, fetch=self.fetch,
                             landscape_broker=self.landscape_broker)
        # The single origin and the landscape-client install were added
        # by the first config-changed hook execution.
        self.assertEqual(
            [("ppa:landscape-client/stable", None)], self.fetch.sources)
        self.assertEqual(1, len(self.fetch.installs))
        [install] = self.fetch.installs
        self.assertEqual(["landscape-client"], install.packages)

    def test_update_origin_with_key(self):
        """
        If the origin contains a key, the key will be added together
        with the apt source.
        """
        config = self.config()
        config.update({"origin": "ppa:landscape-client/stable|key-id"})
        hooks.config_changed(hookenv=self.hookenv, fetch=self.fetch,
                             landscape_broker=self.landscape_broker)
        self.assertEqual(
            [("ppa:landscape-client/stable", "key-id")], self.fetch.sources)

    def test_update_origin_with_account_name(self):
        """
        If the account-name is set and the client already registered,
        updating the origin will result in the apt source being added
        and landscape-client installed/upgraded.
        """
        config = self.config()
        config.update({
            "account-name": "example",
            "origin": "ppa:landscape-client/stable"})
        hooks.config_changed(hookenv=self.hookenv, fetch=self.fetch,
                             landscape_broker=self.landscape_broker)
        self.assertEqual(
            [("ppa:landscape-client/stable", None)], self.fetch.sources)
        [install] = self.fetch.installs
        self.assertEqual(["landscape-client"], install.packages)


class RegistrationRelationJoinedTest(HookTestCase):
    """Test for the registration-relation-joined hook."""

    def setUp(self):
        super(RegistrationRelationJoinedTest, self).setUp()
        self.landscape_broker.config.account_name = "account1"
        self.landscape_broker.config.computer_title = "title"
        self.touch_juju_info_file()
        self.addCleanup(setattr, hooks, "CERT_FILE", hooks.CERT_FILE)
        tmp_dir = self.mkdtemp()
        hooks.CERT_FILE = "%s/cert" % tmp_dir
        self.config()

    def test_ssl_cert_decoded(self):
        """
        A base64 encoded SSL cert in the registration relation is decoded.
        """
        self.juju_broker.commands["relation-get"] = {
            "ssl-public-key": "base64:%s" % "foo".encode("base64"),
            "account-name": "standalone",
            "ping-url": "http://127.0.0.1/ping"}
        hooks.registration_relation(juju_broker=self.juju_broker,
                                    hookenv=self.hookenv, fetch=self.fetch,
                                    landscape_broker=self.landscape_broker)
        self.assertEqual(
            hooks.CERT_FILE, self.landscape_broker.config.ssl_public_key)
        with open(hooks.CERT_FILE) as fh:
            self.assertEqual("foo", fh.read())

    def test_cert_file_in_place(self):
        """
        An ssl-public-key setting which isn't base64 is transparently set.
        """
        self.juju_broker.commands["relation-get"] = {
            "ssl-public-key": "/etc/ssl/my-cert.pem",
            "account-name": "standalone",
            "ping-url": "http://127.0.0.1/ping"}
        hooks.registration_relation(juju_broker=self.juju_broker,
                                    hookenv=self.hookenv, fetch=self.fetch,
                                    landscape_broker=self.landscape_broker)
        self.assertEqual("/etc/ssl/my-cert.pem",
                         self.landscape_broker.config.ssl_public_key)
